#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
from asn1crypto import ocsp as aocsp
from cryptography.x509 import ocsp
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, utils
from cryptography.hazmat.primitives.serialization import pkcs12


def main():
    print("*" * 20, "req")
    data = open("t-ocsp-req.bin", "rb").read()
    ocspr = aocsp.OCSPRequest.load(data)
    print(ocspr.debug())

    print("*" * 20, "resp")
    data = open("t-ocsp-resp.bin", "rb").read()
    ocspresp = aocsp.OCSPResponse.load(data)
    print(ocspresp.debug())


main()



import os
import sys
sys.path.insert(0, os.path.abspath('..'))
# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'endesive'
copyright = '2025, Grzegorz Makarewicz'
author = 'Grzegorz Makarewicz'
#release = '1.0.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
]

templates_path = ['_templates']
exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
    'examples',
    'tests',
    'endesive/pdf/fpdf*',
    'endesive/pdf/PyPDF2*',
    'endesive/pdf/PyPDF2_annotate*',
]



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



#!/usr/bin/env vpython3
import os
import io
from datetime import datetime, timezone, timedelta
from http import server, HTTPStatus
import socketserver
import logging

from cryptography import x509
from cryptography.x509 import ocsp
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization, hashes

logging.basicConfig(level=logging.DEBUG)

log = logging.getLogger(__name__)


class HTTPError(Exception):
    def __init__(self, status: HTTPStatus):
        self.status = status

class HTTPD(socketserver.TCPServer):
    allow_reuse_address = True

    def __init__(self, port):
        super().__init__(("0.0.0.0", port), Handler)


class Handler(server.SimpleHTTPRequestHandler):
    def setup(self):
        super().setup()
        self.extensions_map.update(
            {
                ".wasm": "application/wasm",
            }
        )
        self._root_cert = self.load_cert('demo2_ca.root.crt.pem')
        self._root_key = self.load_key('demo2_ca.root.key.pem', '1234')
        self._sub_cert = self.load_cert('demo2_ca.sub.crt.pem')
        self._sub_key = self.load_key('demo2_ca.sub.key.pem', '1234')

    def xguess_type(self, path):
        return super().guess_type(path)

    def load_cert(self, fname: str) -> x509.Certificate:
        with open(os.path.join("ca", fname), "rb") as f:
            return x509.load_pem_x509_certificate(f.read(), default_backend())

    def load_key(self, fname: str, password: str) -> rsa.RSAPrivateKey:
        with open(os.path.join("ca", fname), "rb") as f:
            key = serialization.load_pem_private_key(
                f.read(),
                password.encode("utf-8"),
                default_backend()
            )
            return key

    def do_ERROR(self, ispost=False):
        print("*" * 20, self.requestline)
        print(self.headers)

        if ispost:
            length = int(self.headers["Content-Length"])
            data = self.rfile.read(length)
            print(data)

        self.send_response(404)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def validate(self, serial: int) -> tuple[x509.Certificate, ocsp.OCSPCertStatus, datetime|None, x509.ReasonFlags|None]:
        try:
            cert = self.load_cert(str(serial))
        except Exception as e:
            print(f"Could not retrieve certificate with serial {serial}: {e}")
            raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR)
        now = datetime.now(tz=timezone.utc)
        if now < cert.not_valid_before_utc:
            return (cert, ocsp.OCSPCertStatus.REVOKED, cert.not_valid_before, x509.ReasonFlags.certificate_hold)
        if cert.not_valid_after_utc < now:
            return (cert, ocsp.OCSPCertStatus.REVOKED, cert.not_valid_after, x509.ReasonFlags.certificate_hold)
        return (cert, ocsp.OCSPCertStatus.GOOD, None, None)

    def _build_ocsp_response(self, req: ocsp.OCSPRequest) -> ocsp.OCSPResponse:
        serial = req.serial_number
        try:
            cert, certificate_status, revocation_time, revocation_reason = self.validate(serial)
        except Exception as e:
            print(f"Could not determine certificate status: %{e}")
            raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR)

        issuer = self.load_cert('demo2_ca.sub.crt.pem')
        if cert.issuer != issuer.subject:
            certificate_status=ocsp.OCSPCertStatus.UNKNOWN

        # Build the response
        time = datetime.now(tz=timezone.utc)-timedelta(days=3)
        builder = ocsp.OCSPResponseBuilder()
        builder = builder.add_response(
            cert=cert,
            issuer=issuer,
            algorithm=hashes.SHA256(),
            cert_status=certificate_status,
            this_update=time,
            next_update=time + timedelta(days=7),
            revocation_time=revocation_time,
            revocation_reason=revocation_reason
        )
        for ext in req.extensions:
            if ext.oid == x509.OCSPNonce.oid:
                builder.add_extension(x509.OCSPNonce(ext.value.nonce), ext.critical)
            elif ext.critical:
                print(f"Could not parse unknown critical extension: {ext}")
                raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR)
        builder = builder.responder_id(
            ocsp.OCSPResponderEncoding.HASH,
            self._sub_cert
        ).sign(
            self._sub_key, hashes.SHA256()
        )
        result = builder.public_bytes(serialization.Encoding.DER)
        return result

    def do_POST_ocsp(self):
        ctype = self.headers["Content-Type"]
        assert ctype == "application/ocsp-request"

        length = int(self.headers["Content-Length"])
        data = self.rfile.read(length)
        ocsprequest = ocsp.load_der_ocsp_request(data)
        ocspresponse = self._build_ocsp_response(ocsprequest)

        f = io.BytesIO(ocspresponse)

        self.send_response(HTTPStatus.OK)
        self.send_header("Content-type", "application/ocsp-response")
        self.send_header("Content-Length", str(len(ocspresponse)))
        self.send_header("Last-Modified", self.date_time_string())
        self.end_headers()
        self.copyfile(f, self.wfile)

    def do_POST(self):
        print("*" * 20, self.requestline)
        print(self.headers)

        try:
            if self.path == "/":
                return self.do_POST_ocsp()
            elif self.path == "/ocsp":
                return self.do_POST_ocsp()
        except:
            import traceback
            traceback.print_exc()
        self.do_ERROR()

        # Extract and print the contents of the POST
        length = int(self.headers["Content-Length"])
        data = self.rfile.read(length)
        print("*" * 10, "data")
        print(data)

    def do_GET_ocsp(self):
        f = open("ca/demo2_ca.root.crt.pem.cer", "rb")
        fs = os.fstat(f.fileno())
        self.send_response(HTTPStatus.OK)
        self.send_header("Content-type", "application/pkix-cert")
        self.send_header("Content-Length", str(fs[6]))
        self.send_header("Last-Modified", self.date_time_string(fs.st_mtime))
        self.end_headers()
        try:
            self.copyfile(f, self.wfile)
        finally:
            f.close()
        return

    def do_GET_crl(self):
        today = datetime.now(tz=timezone.utc)-timedelta(days=3)
        builder = (
            x509.CertificateRevocationListBuilder()
            .last_update(today)
            .next_update(today + timedelta(days=7))
            .issuer_name(self._root_cert.issuer)
        )
        cert_revocation_list = builder.sign(
            private_key=self._root_key,
            algorithm=hashes.SHA256(),
            backend=default_backend(),
        )
        crl = cert_revocation_list.public_bytes(encoding=serialization.Encoding.PEM)
        f = io.BytesIO(crl)

        self.send_response(HTTPStatus.OK)
        self.send_header("Content-type", "application/pkix-crl")
        self.send_header("Content-Length", str(len(crl)))
        self.send_header("Last-Modified", self.date_time_string())
        self.end_headers()
        self.copyfile(f, self.wfile)

    def do_GET(self):
        print("*" * 20, self.requestline)
        print(self.headers)

        try:
            if self.path == "/ocsp":
                return self.do_GET_ocsp()
            if self.path == "/crl":
                return self.do_GET_crl()
        except:
            import traceback
            traceback.print_exc()
        self.do_ERROR()

    def xx(self):
        # import pdb; pdb.set_trace()
        DUMMY_RESPONSE = """\
<html>
<head>
<title>Python Test</title>
</head>

<body>
Test page...success.
<script>
debugger;
</script>
</body>
</html>
"""
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.send_header("Content-length", len(DUMMY_RESPONSE))
        self.end_headers()
        self.wfile.write(DUMMY_RESPONSE)


import threading
import sys
import getopt
import ssl


def main():
    PORT = 33150

    def done():
        input("enter to stop")
        httpd.shutdown()

    print("serving at port", PORT)
    httpd = HTTPD(PORT)
    t = threading.Thread(target=done)
    t.start()
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass
    except:
        import traceback

        traceback.print_exc()


main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import unittest
import typing
import datetime
import os, os.path
import sys
import glob
import uuid

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa, ec
from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.x509.oid import NameOID

tests_root = os.path.dirname(__file__)
fixtures_dir = os.path.join(tests_root, 'fixtures')


def fixture(fname):
    return os.path.join(fixtures_dir, fname)

(
    ca_root_cert,
    ca_root_key,
    ca_sub_cert,
    ca_sub_key,
    cert1_csr,
    cert1_cert,
    cert1_key,
    cert1_pub,
    cert1_p12,
    cert2_csr,
    cert2_cert,
    cert2_key,
    cert2_pub,
    cert2_p12,
    cert3_csr,
    cert3_cert,
    cert3_key,
    cert3_pub,
    cert3_p12,
) = (
    fixture("demo2_ca.root.crt.pem"),
    fixture("demo2_ca.root.key.pem"),
    fixture("demo2_ca.sub.crt.pem"),
    fixture("demo2_ca.sub.key.pem"),
    fixture("demo2_user1.csr.pem"),
    fixture("demo2_user1.crt.pem"),
    fixture("demo2_user1.key.pem"),
    fixture("demo2_user1.pub.pem"),
    fixture("demo2_user1.p12"),
    fixture("demo2_user2.csr.pem"),
    fixture("demo2_user2.crt.pem"),
    fixture("demo2_user2.key.pem"),
    fixture("demo2_user2.pub.pem"),
    fixture("demo2_user2.p12"),
    fixture("demo2_user3.csr.pem"),
    fixture("demo2_user3.crt.pem"),
    fixture("demo2_user3.key.pem"),
    fixture("demo2_user3.pub.pem"),
    fixture("demo2_user3.p12"),
)


class CA(object):
    def __init__(self):
        self.CreateCA()

    def key_create_rsa(self) -> rsa.RSAPrivateKey:
        return rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )

    def key_create_ec(self) -> ec.EllipticCurvePrivateKey:
        return ec.generate_private_key(
            ec.SECP256R1(), default_backend()
        )

    def key_create(self) -> rsa.RSAPrivateKey:
        return rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )

    def key_save(self, fname: str, key: rsa.RSAPrivateKey, password: str) -> None:
        # Write our key to disk for safe keeping
        with open(os.path.join("ca", fname), "wb") as f:
            if not password:
                f.write(
                    key.public_key().public_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PublicFormat.SubjectPublicKeyInfo,
                    )
                )
            else:
                f.write(
                    key.private_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PrivateFormat.TraditionalOpenSSL,
                        encryption_algorithm=serialization.BestAvailableEncryption(
                            password.encode("utf-8")
                        ),
                    )
                )

    def key_load(self, fname: str, password: str) -> rsa.RSAPrivateKey:
        with open(os.path.join("ca", fname), "rb") as f:
            private_key = serialization.load_pem_private_key(
                f.read(), password.encode("utf-8"), default_backend()
            )
            return private_key

    def cert_save(self, fname: str, data: x509.Certificate) -> None:
        with open(os.path.join("ca", fname), "wb") as f:
            f.write(data.public_bytes(serialization.Encoding.PEM))
        with open(os.path.join("ca", fname) + ".cer", "wb") as f:
            f.write(data.public_bytes(serialization.Encoding.DER))

    def cert_load(self, fname: str) -> x509.Certificate:
        with open(os.path.join("ca", fname), "rb") as f:
            return x509.load_pem_x509_certificate(f.read(), default_backend())

    def csr_save(self, fname: str, csr: x509.CertificateSigningRequest) -> None:
        with open(fname, "wb") as f:
            f.write(csr.public_bytes(serialization.Encoding.PEM))

    def csr_load(self, fname: str) -> x509.CertificateSigningRequest:
        with open(os.path.join("ca", fname), "rb") as f:
            return x509.load_pem_x509_csr(data=f.read(), backend=default_backend())

    def csr_create(
        self,
        email: str,
        key: rsa.RSAPrivateKey,
        country: typing.Union[str, None] = None,
        state: typing.Union[str, None] = None,
        locality: typing.Union[str, None] = None,
        organization: typing.Union[str, None] = None,
        commonname: typing.Union[str, None] = None,
    ) -> x509.CertificateSigningRequest:
        names = []
        for t, v in (
            (NameOID.COUNTRY_NAME, country),
            (NameOID.STATE_OR_PROVINCE_NAME, state),
            (NameOID.LOCALITY_NAME, locality),
            (NameOID.ORGANIZATION_NAME, organization),
            (NameOID.COMMON_NAME, commonname),
        ):
            if v:
                names.append(x509.NameAttribute(t, v))
        names.append(x509.NameAttribute(NameOID.EMAIL_ADDRESS, email))
        return (
            x509.CertificateSigningRequestBuilder()
            .subject_name(x509.Name(names))
            .sign(
                # Sign the CSR with our private key.
                key,
                hashes.SHA256(),
                default_backend(),
            )
        )

    def csr_sign(self, fname: str) -> x509.Certificate:
        csr = self.csr_load(fname)
        emails = csr.subject.get_attributes_for_oid(NameOID.EMAIL_ADDRESS)
        names = [
            x509.RFC822Name(emails[0].value),
            #x509.OtherName(
            #    x509.ObjectIdentifier('1.3.6.1.4.1.311.20.2.3'),
            #    #'john.doe@domain.tld'.encode("utf-8")
            #    UTF8String('john.doe@domain.tld').dump()
            #),
            #x509.DNSName('trisoft.com.pl'),
        ]
        return (
            x509.CertificateBuilder()
            .subject_name(csr.subject)
            .issuer_name(self.ca_sub_cert.subject)
            .public_key(csr.public_key())
            .serial_number(uuid.uuid4().int)  # pylint: disable=no-member
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=365))
            .add_extension(
                x509.BasicConstraints(ca=False, path_length=None),
                critical=True
            ).add_extension(
                x509.AuthorityKeyIdentifier.from_issuer_public_key(
                    self.ca_sub_pk.public_key()
                ),
                critical=False,
            ).add_extension(
                x509.CRLDistributionPoints(
                    [
                        x509.DistributionPoint(
                            full_name=[
                                x509.UniformResourceIdentifier(
                                    "http://ca.trisoft.com.pl/crl"
                                )
                            ],
                            relative_name=None,
                            reasons=None,
                            crl_issuer=None,
                        )
                    ]
                ),
                critical=False,
            ).add_extension(
                x509.AuthorityInformationAccess(
                    [
                        x509.AccessDescription(
                            x509.OID_CA_ISSUERS,
                            x509.UniformResourceIdentifier(
                                "http://ca.trisoft.com.pl/cacert"
                            ),
                        ),
                        x509.AccessDescription(
                            x509.OID_OCSP,
                            x509.UniformResourceIdentifier(
                                "http://ca.trisoft.com.pl/ocsp"
                            ),
                        ),
                    ]
                ),
                critical=False,
            ).add_extension(
                x509.SubjectAlternativeName(names),
                critical=False,
            ).add_extension(
                # certificate_policies
                x509.ExtendedKeyUsage([
                    x509.OID_CLIENT_AUTH, # 1.3.6.1.5.5.7.3.2
                    #x509.OID_SERVER_AUTH,
                    x509.OID_EMAIL_PROTECTION,
                    x509.ObjectIdentifier("1.3.6.1.4.1.311.10.3.12"), # document signing
                    x509.ObjectIdentifier("1.3.6.1.5.5.7.3.36"), # document signing
                    #x509.ObjectIdentifier("1.3.6.1.5.5.7.3.21"), # ssh client
                    #x509.ObjectIdentifier("1.3.6.1.5.5.7.3.22"), # ssh server
                    #1.2.840.113583.1.1.7.1.0 .. 11 # https://www.adobe.com/devnet-docs/acrobatetk/tools/DigSigDC/oids.html
                ]),
                critical=False,
            ).add_extension(
                x509.SubjectKeyIdentifier.from_public_key(csr.public_key()),
                critical=False,
            ).add_extension(
                x509.KeyUsage(
                    # Digital Signature: Indicates that the key can be used for digital signatures to verify the authenticity and integrity of data. 
                    digital_signature=True,
                    # Non-Repudiation: Used in conjunction with digital signatures to provide an additional layer of protection against denial of signature. 
                    content_commitment=True,  # nonRepudiation
                    # Key Encipherment: Specifies that the key can be used for encrypting other keys, typically for key transport. 
                    key_encipherment=True,
                    # Data Encipherment: Indicates that the key can be used for data encryption and decryption. 
                    data_encipherment=True,
                    # Key Agreement: Used when the key is involved in key exchange agreements, such as Diffie-Hellman. 
                    key_agreement=True,
                    # Encipher Only: Specifies that the key can only be used for encryption, not decryption. 
                    encipher_only=False,
                    # Decipher Only: Specifies that the key can only be used for decryption, not encryption. 
                    decipher_only=False,
                    # ca
                    # Certificate Signing: Specifies that the key can be used to sign other certificates, typically used by Certificate Authorities. 
                    key_cert_sign=False,
                    # CRL Signing: Indicates that the key can be used to sign Certificate Revocation Lists (CRLs). 
                    crl_sign=False,
                ),
                critical=True,
            ).sign(
                private_key=self.ca_sub_pk,
                algorithm=hashes.SHA256(),
                backend=default_backend(),
            )
        )

    def pk12_save(
        self,
        name: bytes,
        cert: x509.Certificate,
        key: rsa.RSAPrivateKey,
        fname: str,
        password: str,
    ) -> None:
        data = pkcs12.serialize_key_and_certificates(
            name=name,
            key=key,
            cert=cert,
            cas=[self.ca_sub_cert],
            encryption_algorithm=serialization.BestAvailableEncryption(
                password.encode("utf8")
            ),
        )
        with open(os.path.join("ca", fname), "wb") as f:
            f.write(data)

    def pk12_load(self, fname: str, password: str) -> pkcs12.PKCS12KeyAndCertificates:
        with open(os.path.join("ca", fname), "rb") as fp:
            return pkcs12.load_key_and_certificates(
                fp.read(), password.encode("utf-8"), default_backend()
            )

    def ca_createroot(self, key: rsa.RSAPrivateKey) -> x509.Certificate:
        subject = issuer = x509.Name(
            [
            x509.NameAttribute(NameOID.COMMON_NAME, "AA TriSoft Root CA"),
            ]
        )
        return (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer)
            .public_key(key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(
                # Our certificate will be valid for 40 years
                datetime.datetime.utcnow()
                + datetime.timedelta(days=40 * 365)
            ).add_extension(
                x509.BasicConstraints(
                    ca=True,
                    path_length=None,  # pathlen: is equal to the number of CAs/ICAs it can sign
                ),
                critical=True,
            ).add_extension(
                x509.AuthorityKeyIdentifier.from_issuer_public_key(key.public_key()),
                critical=False,
            ).add_extension(
                x509.SubjectKeyIdentifier.from_public_key(key.public_key()),
                critical=False,
            ).add_extension(
                x509.KeyUsage(
                    digital_signature=True,
                    content_commitment=False,  # nonRepudiation
                    key_encipherment=False,
                    data_encipherment=False,
                    key_agreement=False,
                    encipher_only=False,
                    decipher_only=False,
                    # ca
                    key_cert_sign=True,
                    crl_sign=True,
                ),
                critical=True,
            ).sign(
                # Sign our certificate with our private key
                key,
                hashes.SHA256(),
                default_backend(),
            )
        )

    def ca_createsub(self, key: rsa.RSAPrivateKey, rootcert: x509.Certificate, rootkey: rsa.RSAPrivateKey) -> x509.Certificate:
        subject = x509.Name(
            [
            x509.NameAttribute(NameOID.COMMON_NAME, "AA TriSoft Intermediate CA"),
            ]
        )
        return (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(rootcert.subject)
            .public_key(key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(
                # Our certificate will be valid for 10 years
                datetime.datetime.utcnow()
                + datetime.timedelta(days=10 * 365)
            ).add_extension(
                x509.BasicConstraints(
                    ca=True,
                    path_length=0,  # pathlen: is equal to the number of CAs/ICAs it can sign
                ),
                critical=True,
            ).add_extension(
                x509.CRLDistributionPoints(
                    [
                        x509.DistributionPoint(
                            full_name=[
                                x509.UniformResourceIdentifier(
                                    "http://ca.trisoft.com.pl/crl"
                                )
                            ],
                            relative_name=None,
                            reasons=None,
                            crl_issuer=None,
                        )
                    ]
                ),
                critical=False,
            ).add_extension(
                x509.AuthorityInformationAccess(
                    [
                        x509.AccessDescription(
                            x509.OID_CA_ISSUERS,
                            x509.UniformResourceIdentifier(
                                "http://ca.trisoft.com.pl/cacert"
                            ),
                        ),
                        x509.AccessDescription(
                            x509.OID_OCSP,
                            x509.UniformResourceIdentifier(
                                "http://ca.trisoft.com.pl/ocsp"
                            ),
                        ),
                    ]
                ),
                critical=False,
            ).add_extension(
                x509.AuthorityKeyIdentifier.from_issuer_subject_key_identifier(
                    rootcert.extensions.get_extension_for_class(x509.SubjectKeyIdentifier).value
                ),
                critical=False,
            ).add_extension(
                x509.SubjectKeyIdentifier.from_public_key(key.public_key()),
                critical=False,
            ).add_extension(
                x509.KeyUsage(
                    digital_signature=True,
                    content_commitment=False,  # nonRepudiation
                    key_encipherment=False,
                    data_encipherment=False,
                    key_agreement=False,
                    encipher_only=False,
                    decipher_only=False,
                    # ca
                    key_cert_sign=True,
                    crl_sign=True,
                ),
                critical=True,
            ).sign(
                # Sign our certificate with our private key
                rootkey,
                hashes.SHA256(),
                default_backend(),
            )
        )

    def CreateCA(self) -> None:
        create = not (
            os.path.exists(os.path.join("ca", ca_sub_cert))
            and os.path.exists(os.path.join("ca", ca_sub_key))
        )
        if create:
            for fqname in glob.glob("ca/*"):
                os.unlink(fqname)
            ca_root_pk = self.key_create()
            ca_root_cr = self.ca_createroot(ca_root_pk)

            self.key_save(ca_root_key, ca_root_pk, "1234")
            self.cert_save(ca_root_cert, ca_root_cr)

            ca_sub_pk = self.key_create()
            ca_sub_cr = self.ca_createsub(ca_sub_pk, ca_root_cr, ca_root_pk)

            self.key_save(ca_sub_key, ca_sub_pk, "1234")
            self.cert_save(ca_sub_cert, ca_sub_cr)

            with open(ca_sub_cert, 'rb') as fp:
                data = fp.read()
            with open(ca_root_cert, 'rb') as fp:
                data += fp.read()
            with open(fixture('root.pem'), 'wb') as fp:
                fp.write(data)
        else:
            ca_root_pk = self.key_load(ca_root_key, "1234")
            ca_root_cr = self.cert_load(ca_root_cert)
            ca_sub_pk = self.key_load(ca_sub_key, "1234")
            ca_sub_cr = self.cert_load(ca_sub_cert)

        self.ca_root_cert = ca_root_cr
        self.ca_root_pk = ca_root_pk
        self.ca_sub_cert = ca_sub_cr
        self.ca_sub_pk = ca_sub_pk

    def USER(self, no, cert, cert_key, cert_pub, cert_p12) -> None:
        create = not (
            os.path.exists(os.path.join("ca", cert))
            and os.path.exists(os.path.join("ca", cert_key))
            and os.path.exists(os.path.join("ca", cert_pub))
            and os.path.exists(os.path.join("ca", cert_p12))
        )
        if create:
            client_pk = self.key_create()
            client_csr = self.csr_create(
                "demo%d@trisoft.com.pl" % no,
                client_pk,
                commonname='trisoft.com.pl',
            )
            client_cert = self.csr_sign(client_csr)
            self.cert_save(cert, client_cert)
            self.key_save(cert_key, client_pk, "1234")
            self.key_save(cert_pub, client_pk, None)
            self.pk12_save(
                "USER cert".encode("utf-8"), client_cert, client_pk, cert_p12, "1234"
            )

            # os.chmod(pkcs12 % user, stat.S_IRUSR | stat.S_IWUSR)  # 0o600 perms

    def USERs(self):
        self.USER(1, cert1, cert1_key, cert1_pub, cert1_p12)
        self.USER(2, cert2, cert2_key, cert2_pub, cert2_p12)
        self.USER(3, cert3, cert3_key, cert3_pub, cert3_p12)

    def USER(self, no, cert, cert_key, cert_pub, cert_p12) -> None:
        if no == 1:
            client_pk = self.key_create_rsa()
            fname = cert1_csr
        elif no == 2:
            client_pk = self.key_create_rsa()
            fname = cert2_csr
        else:
            client_pk = self.key_create_ec()
            fname = cert3_csr
        client_csr = self.csr_create("demo%d@trisoft.com.pl" % no, client_pk, commonname="USER %s"%no)
        self.csr_save(fname, client_csr)
        client_cert = self.csr_sign(fname)
        self.cert_save(cert, client_cert)
        self.key_save(cert_key, client_pk, "1234")
        self.key_save(cert_pub, client_pk, None)
        self.pk12_save(
            "USER cert".encode("utf-8"), client_cert, client_pk, cert_p12, "1234"
        )

    def createUSERS(self):
        self.USER(1, cert1_cert, cert1_key, cert1_pub, cert1_p12)
        self.USER(2, cert2_cert, cert2_key, cert2_pub, cert2_p12)
        self.USER(3, cert3_cert, cert3_key, cert3_pub, cert3_p12)


class CATests(unittest.TestCase):
    def test_create_ca(self):
        cls = CA()

    def test_create_user(self):
        cls = CA()
        cls.createUSERS()

if __name__ == '__main__':
    cls = CATests()
    for n in dir(cls):
        if n.split('_')[0] == 'test':
            print(n)
            try:
                getattr(cls, n)()
            except Exception as exc:
                pass



#!/usr/bin/env vpython3

""" A tool which takes in pdf, pfx file, password as input and gives out a corresponding signed pdf

"""
import argparse
import re
import sys
import datetime

from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.x509.oid import NameOID
from endesive import pdf

signature_string = lambda organization, date, country : (organization + '\nDATE: '+ date)

def eprint(error):
    print(error, file=sys.stderr)

def load_pfx(file_path, password):
    """ Function to load pkcs12 object from the given password protected pfx file."""

    with open(file_path, 'rb') as fp:
        return pkcs12.load_key_and_certificates(fp.read(), password.encode(), backends.default_backend())

def create_args():
    """Creates CLI arguments for the pdfSigner script."""

    parser = argparse.ArgumentParser(description='Script for digitally signing a pdf')
    parser.add_argument('pfx_certificate', type=str, help='Specify keystore file in .pfx format (Mandatory)')
    parser.add_argument('password', type=str, help=' Specify password for keystore file (mandatory)')
    parser.add_argument('src', type=str,
        help='Specify the source file (.pdf) that needs to be digitally signed. Only 1 file at a time can be signed. (Mandatory) ')
    parser.add_argument('-d', '--dest', type=str,
        help='Specify the destination file where digitally signed content will be stored.When not specified, by default it will '
        'digitally sign the source file.(Mandatory) \n'
        'E.g. Given source file /var/hp/some.pdf will be digitally signed')
    parser.add_argument('-c', '--coords', type=str,
        help='Specify the co-ordinates of where you want the digital signature to be placed on the PDF file page.(Optional)\n'
        'Format: Accepts 4 comma-separated float values (without spaces). E.g. 1,2,3,4 ')
    parser.add_argument('-p', '--page', type=int,
        help='You can specify the page number of PDF file where digital signature(Optional)')

    return parser.parse_args()

def validate_args(args):
    """Validating commandline arguments raises valueError exception with if any command
    line arguments are not valid."""

    IS_PFX = lambda pfx_certificate: re.match( r'^(.[^,]+)(.pfx|.PFX|.p12|.P12){1}$', pfx_certificate)
    if not IS_PFX(args.pfx_certificate):
        raise ValueError('Not a proper pfx file with .pfx or .PFX extension')
    if args.coords:
        for num in args.coords.split(','):
            if not num.isdigit():
                raise ValueError('Coords are not integers')


OID_NAMES = {
    NameOID.COMMON_NAME: 'CN',
    NameOID.COUNTRY_NAME: 'C',
    NameOID.DOMAIN_COMPONENT: 'DC',
    NameOID.EMAIL_ADDRESS: 'E',
    NameOID.GIVEN_NAME: 'G',
    NameOID.LOCALITY_NAME: 'L',
    NameOID.ORGANIZATION_NAME: 'O',
    NameOID.ORGANIZATIONAL_UNIT_NAME: 'OU',
    NameOID.SURNAME: 'SN'
}
def get_rdns_names(rdns):
    names = {}
    for oid in OID_NAMES:
        names[OID_NAMES[oid]] = ''
    for rdn in rdns:
        for attr in rdn._attributes:
            if attr.oid in OID_NAMES:
                names[OID_NAMES[attr.oid]] = attr.value
    return names

def run():
    args = create_args()

    try:
        validate_args(args)
    except ValueError as e:
        import traceback; traceback.print_exc()
        sys.exit(1)

    try:
        # Load the PKCS12 object from the pfx file
        p12pk, p12pc, p12oc = load_pfx(args.pfx_certificate, args.password)

        names = get_rdns_names(p12pc.subject.rdns)
        #default coords of bottom right corner in a pdf page
        coords = [350, 50, 550, 150]
        if args.coords:
          coords = [int(coord) for coord in args.coords.split(',') if coord]
        page = args.page if args.page else 1
        dest = args.dest if args.dest else args.src
        date = datetime.datetime.utcnow() - datetime.timedelta(hours=12)
        date = date.strftime('%Y%m%d%H%M%S+00\'00\'')
        signature = signature_string(names['CN'], date, names['C'])
        dct = {
            'sigflags': 3,
            'sigpage': page - 1,
            'contact': 'finacctind@endurance.com',
            #'location': subject.C.encode(),
            'location': 'Szczecin',
            'signingdate': date,
            'signingdate': '20201119195200+02\'00\'',
            'reason': 'Signed by endurance',
            'signature': signature,
            'signaturebox': tuple(coords[:4]),
        }

        input_file = args.src
        datau = open(input_file, 'rb').read()
        datas = pdf.cms.sign(datau,
                     dct,
                     p12pk, p12pc, p12oc,
                     'sha256'
                     )

        output_file = input_file.replace(input_file, dest)
        with open(output_file, 'wb') as fp:
          fp.write(datau)
          fp.write(datas)
    except Exception as e:
        import traceback; traceback.print_exc()
        eprint(e)
        sys.exit()

if __name__ == '__main__':
    run()



#!/usr/bin/env vpython3
import os
#os.environ['JAVA_HOME'] = '/usr/lib/jvm/msopenjdk-21-amd64/'
os.environ['JVM_PATH'] = '/usr/lib/jvm/msopenjdk-21-amd64/lib/server/libjvm.so'
import sys
import jnius_config

PATH=['.']
def walk(top):
    todo = []
    for fname in os.listdir(top):
        fqname = os.path.join(top, fname)
        if os.path.isdir(fqname):
            todo.append(fqname)
        elif fname.split('.')[-1] == 'jar':
            PATH.append(fqname)
    for fqname in todo:
        walk(fqname)

walk('/devel/lib/java/pdfbox')
jnius_config.set_classpath(*PATH)


import jnius



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
from endesive import pdf, hsm

import os
import sys
import sysconfig
import datetime
from cryptography import x509
from cryptography.hazmat import backends

os.environ['SOFTHSM2_CONF'] = 'softhsm2.conf'
if not os.path.exists(os.path.join(os.getcwd(), 'softhsm2.conf')):
    open('softhsm2.conf', 'wt').write('''\
log.level = DEBUG
directories.tokendir = %s/softhsm2/
objectstore.backend = file
slots.removable = false
''' % os.getcwd())
if not os.path.exists(os.path.join(os.getcwd(), 'softhsm2')):
    os.mkdir(os.path.join(os.getcwd(), 'softhsm2'))

#
#!/bin/bash
#SOFTHSM2_CONF=softhsm2.conf
#softhsm2-util --label "endesive" --slot 1 --init-token --pin secret1 --so-pin secret2
#softhsm2-util --show-slots
#

if sys.platform == 'win32':
    dllpath = r'W:\binw\SoftHSM2\lib\softhsm2-x64.dll'
else:
    dllpath = os.path.join(sysconfig.get_config_var('LIBDIR'), "softhsm/libsofthsm2.so")

import PyKCS11 as PK11

class Signer(hsm.HSM):
    def certificate(self):
        self.login("endesieve", "secret1")
        keyid = bytes((0x66,0x66,0x90))
        try:
            pk11objects = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_CERTIFICATE)])
            all_attributes = [
                #PK11.CKA_SUBJECT,
                PK11.CKA_VALUE,
                #PK11.CKA_ISSUER,
                #PK11.CKA_CERTIFICATE_CATEGORY,
                #PK11.CKA_END_DATE,
                PK11.CKA_ID,
            ]

            for pk11object in pk11objects:
                try:
                    attributes = self.session.getAttributeValue(pk11object, all_attributes)
                except PK11.PyKCS11Error as e:
                    continue

                attrDict = dict(list(zip(all_attributes, attributes)))
                cert = bytes(attrDict[PK11.CKA_VALUE])
                if keyid == bytes(attrDict[PK11.CKA_ID]):
                    return keyid, cert
        finally:
            self.logout()
        return None, None

    def sign(self, keyid, data, mech):
        self.login("endesieve", "secret1")
        try:
            privKey = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_PRIVATE_KEY), (PK11.CKA_ID, keyid)])[0]
            mech = getattr(PK11, 'CKM_%s_RSA_PKCS' % mech.upper())
            sig = self.session.sign(privKey, data, PK11.Mechanism(mech, None))
            return bytes(sig)
        finally:
            self.logout()

def main():
    tspurl = "http://time.certum.pl"
    tspurl = "http://public-qlts.certum.pl/qts-17"

    ocspurl = 'https://ocsp.certum.pl/'
    ocspissuer = open('CertumDigitalIdentificationCASHA2.crt', 'rb').read()
    ocspissuer = x509.load_pem_x509_certificate(ocspissuer, backends.default_backend())

    date = datetime.datetime.utcnow() - datetime.timedelta(hours=12)
    date = date.strftime('D:%Y%m%d%H%M%S+00\'00\'')
    dct = {
        'sigflags': 3,
        'contact': 'mak@trisoft.com.pl',
        'location': 'Szczecin',
        'signingdate': date.encode(),
        'reason': 'Dokument podpisany cyfrowo',
        'application': 'app:xyz',
    }

    clshsm = Signer(dllpath)

    fname = 'pdf.pdf'
    if len (sys.argv) > 1:
        fname = sys.argv[1]

    datau = open(fname, 'rb').read()
    datas = pdf.cms.sign(datau, dct,
        None, None,
        [],
        'sha256',
        clshsm,
        tspurl,
        ocspurl=ocspurl,
        ocspissuer=ocspissuer
    )
    fname = fname.replace('.pdf', '-signed-cms-hsm.pdf')
    with open(fname, 'wb') as fp:
        fp.write(datau)
        fp.write(datas)


main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import requests
from cryptography import x509
from cryptography.x509 import ocsp
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, utils
from cryptography.hazmat.primitives.serialization import pkcs12

def get_from_cert(cert, what):
    try:
        aia = cert.extensions.get_extension_for_oid(x509.oid.ExtensionOID.AUTHORITY_INFORMATION_ACCESS)
        for access_description in aia.value:
            if access_description.access_method == what:
                return access_description.access_location.value
    except x509.ExtensionNotFound:
        return None
    return None

def get_crl_from_cert(cert):
    try:
        cdp = cert.extensions.get_extension_for_oid(x509.oid.ExtensionOID.CRL_DISTRIBUTION_POINTS)
        return cdp.value[0].full_name[0].value
    except x509.ExtensionNotFound:
        return None
    return None

def main():
    with open("ca/demo2_user1.p12", "rb") as fp:
        p12 = pkcs12.load_key_and_certificates(
            fp.read(), b"1234", backends.default_backend()
        )

    ocspurl = 'http://ca.trisoft.com.pl/'

    ocspissuerurl = get_from_cert(p12[1], x509.OID_OCSP)
    rootcerturl = get_from_cert(p12[1], x509.OID_CA_ISSUERS)
    crlurl = get_crl_from_cert(p12[1])

    print('crlurl', crlurl)
    response = requests.get(crlurl)
    with open('t-ocsp-crl.der', 'wb') as fp:
        fp.write(response.content)

    print('ocspissuerurl', ocspissuerurl)
    response = requests.get(ocspissuerurl)
    ocspissuer = response.content
    with open('t-ocsp-issuer.der', 'wb') as fp:
        fp.write(ocspissuer)

    ocspissuer = x509.load_der_x509_certificate(ocspissuer, backends.default_backend())

    builder = ocsp.OCSPRequestBuilder()
    builder = builder.add_certificate(p12[1], ocspissuer, hashes.SHA1())
    req = builder.build()
    data = req.public_bytes(serialization.Encoding.DER)

    open("t-ocsp-req.bin", "wb").write(data)
    response = requests.post(
        ocspurl,
        headers={"Content-Type": "application/ocsp-request"},
        data=data,
    )
    open("t-ocsp-resp.bin", "wb").write(response.content)


main()



# *-* coding: utf-8 *-*
import os
import glob
import hashlib
import datetime

from asn1crypto import x509, core, pem, cms

import certifi

from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding, ec
from cryptography.x509.verification import PolicyBuilder, Store
from cryptography import x509 as cx509
from cryptography.hazmat.backends import default_backend


class VerifyData(object):
    def __init__(self, trustedCerts=None):
        with open(certifi.where(), "rb") as pems:
            certs = cx509.load_pem_x509_certificates(pems.read())
        if trustedCerts is not None:
            for cert_bytes in trustedCerts:
                certs.append(cx509.load_pem_x509_certificate(cert_bytes))
        #self.trustedCerts = trustedCerts
        store = Store(certs)
        self.verifier = PolicyBuilder(
            ).store(store
            ).time(datetime.datetime.utcnow()
            ).max_chain_depth(4
            ).build_client_verifier()

    def verify(self, datas, datau):
        signed_data = cms.ContentInfo.load(datas)["content"]
        # signed_data.debug()

        signature = signed_data["signer_infos"][0]["signature"].native
        algo = signed_data["digest_algorithms"][0]["algorithm"].native
        attrs = signed_data["signer_infos"][0]["signed_attrs"]
        mdData = getattr(hashlib, algo)(datau).digest()
        if attrs is not None and not isinstance(attrs, core.Void):
            mdSigned = None
            for attr in attrs:
                if attr["type"].native == "message_digest":
                    mdSigned = attr["values"].native[0]
            signedData = attrs.dump()
            signedData = b"\x31" + signedData[1:]
        else:
            mdSigned = mdData
            signedData = datau
        hashok = mdData == mdSigned
        cert = None
        othercerts = []
        serial = signed_data["signer_infos"][0]["sid"].native["serial_number"]
        for pdfcert in signed_data["certificates"]:
            if serial != pdfcert.native["tbs_certificate"]["serial_number"]:
                othercerts.append(
                    cx509.load_pem_x509_certificate(
                        pem.armor("CERTIFICATE", pdfcert.chosen.dump())
                    )
                )
            else:
                assert cert is None
                cert = cx509.load_pem_x509_certificate(
                    pem.armor("CERTIFICATE", pdfcert.chosen.dump())
                )
        public_key = cert.public_key()

        sigalgo = signed_data["signer_infos"][0]["signature_algorithm"]
        # sigalgo.debug()
        sigalgoname = sigalgo.signature_algo
        if isinstance(public_key, ec.EllipticCurvePublicKey):
            try:
                public_key.verify(
                    signature,
                    signedData,
                    ec.ECDSA(getattr(hashes, algo.upper())()),
                )
                signatureok = True
            except Exception as e:
                signatureok = False
        elif sigalgoname == "rsassa_pss":
            parameters = sigalgo["parameters"]
            # parameters.debug()
            # print(parameters.native)
            salgo = parameters["hash_algorithm"].native["algorithm"].upper()
            mgf = getattr(
                padding, parameters["mask_gen_algorithm"].native["algorithm"].upper()
            )(getattr(hashes, salgo)())
            salt_length = parameters["salt_length"].native
            try:
                public_key.verify(
                    signature,
                    signedData,
                    padding.PSS(mgf, salt_length),
                    getattr(hashes, salgo)(),
                )
                signatureok = True
            except:
                signatureok = False
        elif sigalgoname == "rsassa_pkcs1v15":
            try:
                public_key.verify(
                    signature,
                    signedData,
                    padding.PKCS1v15(),
                    getattr(hashes, algo.upper())(),
                )
                signatureok = True
            except:
                signatureok = False
        else:
            raise ValueError("Unknown signature algorithm")

        try:
            self.verifier.verify(cert, othercerts)
            certok = True
        except Exception as ex:
            print("*" * 10, "failed certificate verification:", ex)
            print("cert.issuer:", cert.issuer)
            print("cert.subject:", cert.subject)
            certok = False
        return (hashok, signatureok, certok)


def verify(datas:bytes, datau:bytes, certs:list[x509.Certificate]=None) -> tuple[bool, bool, bool]:
    """
    Verify signed data.

    :param data: Email data as bytes.
    :param certs: List of additional certificates used to verify signature (system independent).
    :return:
        hashok, signatureok, certok

        hashok : bool
            True if the hash matches.
        signatureok : bool
            True if the signature is valid.
        certok : bool
            True if the certificate used for signing is trusted and valid.
    """
    cls = VerifyData(certs)
    return cls.verify(datas, datau)



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import requests
import hashlib
import time
from asn1crypto import cms, algos, core, keys, pem, tsp, x509, ocsp, util


def timestamp(unhashed, hashalgo, url, credentials, req_options, prehashed=None):
    if prehashed:
        hashed_value = prehashed
    else:
        hashed_value = getattr(hashlib, hashalgo)(unhashed).digest()
    tspreq = tsp.TimeStampReq({
        "version": 1,
        "message_imprint": tsp.MessageImprint({
            "hash_algorithm": algos.DigestAlgorithm({'algorithm': hashalgo}),
            "hashed_message": hashed_value,
            }),
        #'req_policy', ObjectIdentifier, {'optional': True}),
        "nonce": int(time.time()*1000),
        "cert_req": True,
        #'extensions': tsp.Extensions()
        })
    tspreq = tspreq.dump()
    open('t-ts-req.bin', 'wb').write(tspreq)

    tspheaders = {"Content-Type": "application/timestamp-query"}
    username = credentials.get("username", None)
    password = credentials.get("password", None)
    if username and password:
        auth_header_value = b64encode(bytes(username + ':' + password, "utf-8")).decode("ascii")
        tspheaders["Authorization"] = f"Basic {auth_header_value}"

    tspresp = requests.post(url, data=tspreq, headers=tspheaders, **req_options)
    if tspresp.headers.get('Content-Type', None) == 'application/timestamp-reply':
        open('t-ts-resp.bin','wb').write(tspresp.content)
        tspresp = tsp.TimeStampResp.load(tspresp.content)

        if tspresp['status']['status'].native == 'granted':
            attrs = [
                cms.CMSAttribute({
                    'type': cms.CMSAttributeType('signature_time_stamp_token'),
                    'values': cms.SetOfContentInfo([
                        cms.ContentInfo({
                            'content_type': cms.ContentType('signed_data'),
                            'content': tspresp["time_stamp_token"]["content"],
                            })
                        ])
                    })
                ]
            return attrs
        else:
            raise ValueError("TimeStampResponse status is not granted")
    else:
        raise ValueError("TimeStampResponse has invalid content type")



def main():
    unhashed = b'ala ma kota'
    hashalgo = 'sha256'
    #url = "http://time.certum.pl"
    url = "http://public-qlts.certum.pl/qts-17"
    #url = 'https://freetsa.org/tsr'
    credentials = {}
    req_options = {}
    prehashed = None
    attrs = timestamp(unhashed, hashalgo, url, credentials, req_options, prehashed)

main()



# *-* coding: utf-8 *-*
import os
import sys
import glob
import logging
import hashlib
import datetime

from asn1crypto import x509, core, pem, cms, tsp, crl, ocsp, pdf

import certifi

from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding, ec
from cryptography.x509 import verification
from cryptography import x509 as cx509
from cryptography.hazmat.backends import default_backend

from endesive import verifier


logger = logging.getLogger(__name__)


class PDFVerifier:
    def __init__(self, pdf_data: bytes, trustedCerts=None, systemCertsPath=None):
        self.pdf_data = pdf_data
        self.modified = False
        self.wholefile = False
        self.byte_ranges = []

        #print('certifi.where()', certifi.where())
        with open(certifi.where(), "rb") as pems:
            certs = cx509.load_pem_x509_certificates(pems.read())
        if trustedCerts is not None:
            for cert_bytes in trustedCerts:
                certs.append(cx509.load_pem_x509_certificate(cert_bytes))
        #self.trustedCerts = trustedCerts

        self.certs = certs
        self.verifier = verification.PolicyBuilder(
            ).store(verification.Store(certs)
            ).time(datetime.datetime.utcnow()
            ).max_chain_depth(4
            ).build_client_verifier()

    def validator_cb(self, policy, cert, ext):
        print('*'*20, 'validator_cb')
        print('pol:', policy)
        print('cert:', cert)
        print('ext:', ext)
        # Any exception from the python validator is treated as failure.
        #raise ValueError("...")

    def validate_certificate(self, cert, othercerts=[]) -> bool:
        try:
            self.verifier.verify(cert, othercerts)
            certok = True
        except Exception as ex:
            logger.exception(ex)
            certok = False
        return certok

    def is_valid_pdf(self) -> bool:
        return b"%PDF-" in self.pdf_data[:1024]

    def is_signed(self) -> bool:
        n = 0
        while True:
            n = self.pdf_data.find(b"/ByteRange", n)
            if n == -1:
                break
            start = self.pdf_data.find(b"[", n)
            stop = self.pdf_data.find(b"]", start)
            if start == -1 or stop == -1:
                self.modified = True
                return False
            n = stop + 1
            try:
                br = [int(i, 10) for i in self.pdf_data[start + 1 : stop].split()]
                assert self.pdf_data[br[1]] == 60 and self.pdf_data[br[2]-1] == 62
            except:
                self.modified = True
                return False
            self.byte_ranges.append(br)

        if len(self.byte_ranges) == 0:
            return False

        byte_range = self.byte_ranges[-1] # last signature
        if byte_range[0]!=0 or byte_range[2]+byte_range[3] != len(self.pdf_data):
            self.wholefile = False
            return False
        self.wholefile = True

        return True

    def decompose_signed_data(self, datau: bytes, signed_data: cms.SignedData) -> tuple:
        signature = signed_data["signer_infos"][0]["signature"].native
        algo = signed_data["digest_algorithms"][0]["algorithm"].native
        attrs = signed_data["signer_infos"][0]["signed_attrs"]
        mdData = getattr(hashlib, algo)(datau).digest()
        if attrs is not None and not isinstance(attrs, core.Void):
            mdSigned = None
            for attr in attrs:
                if attr["type"].native == "message_digest":
                    mdSigned = attr["values"].native[0]
            signedData = attrs.dump()
            signedData = b"\x31" + signedData[1:]
        else:
            mdSigned = mdData
            signedData = datau
        hashok = mdData == mdSigned
        cert = None
        othercerts = []
        serial = signed_data["signer_infos"][0]["sid"].native["serial_number"]
        for pdfcert in signed_data["certificates"]:
            if serial != pdfcert.native["tbs_certificate"]["serial_number"]:
                othercerts.append(
                    cx509.load_pem_x509_certificate(
                        pem.armor("CERTIFICATE", pdfcert.chosen.dump())
                    )
                )
            else:
                assert cert is None
                cert = cx509.load_pem_x509_certificate(
                    pem.armor("CERTIFICATE", pdfcert.chosen.dump())
                )
        public_key = cert.public_key()

        sigalgo = signed_data["signer_infos"][0]["signature_algorithm"]
        sigalgoname = sigalgo.signature_algo
        if isinstance(public_key, ec.EllipticCurvePublicKey):
            try:
                public_key.verify(
                    signature,
                    signedData,
                    ec.ECDSA(getattr(hashes, algo.upper())()),
                )
                signatureok = True
            except Exception as e:
                signatureok = False
        elif sigalgoname == "rsassa_pss":
            parameters = sigalgo["parameters"]
            # parameters.debug()
            # print(parameters.native)
            salgo = parameters["hash_algorithm"].native["algorithm"].upper()
            mgf = getattr(
                padding, parameters["mask_gen_algorithm"].native["algorithm"].upper()
            )(getattr(hashes, salgo)())
            salt_length = parameters["salt_length"].native
            try:
                public_key.verify(
                    signature,
                    signedData,
                    padding.PSS(mgf, salt_length),
                    getattr(hashes, salgo)(),
                )
                signatureok = True
            except:
                signatureok = False
        elif sigalgoname == "rsassa_pkcs1v15":
            try:
                public_key.verify(
                    signature,
                    signedData,
                    padding.PKCS1v15(),
                    getattr(hashes, algo.upper())(),
                )
                signatureok = True
            except:
                signatureok = False
        else:
            raise ValueError("Unknown signature algorithm")

        tspdata = None
        for attr in signed_data["signer_infos"][0]["unsigned_attrs"]:
            if attr["type"].native == "signature_time_stamp_token":
                for v in attr["values"]:
                    if v["content_type"].native == "signed_data":
                        tspdata = v["content"]
        crls = signed_data["crls"]

        #print(signed_data, tspdata, crls, cert, othercerts, hashok, signatureok)
        return (signed_data, tspdata, crls, cert, othercerts, hashok, signatureok)

    def decompose_signature(self) -> tuple:
        byte_range = self.byte_ranges[-1] # last signature
        contents = self.pdf_data[byte_range[0] + byte_range[1] + 1 : byte_range[2] - 1]
        try:
            signaturebytes = bytes.fromhex(contents.decode("utf8"))
        except:
            return False
        data1 = self.pdf_data[byte_range[0] : byte_range[0] + byte_range[1]]
        data2 = self.pdf_data[byte_range[2] : byte_range[2] + byte_range[3]]
        datau = data1 + data2

        signed_data = cms.ContentInfo.load(signaturebytes)["content"]
        return self.decompose_signed_data(datau, signed_data)

    def verify_ocsp_data(self, cert, othercerts, crldata):
        for crl1 in crldata:
            # clr1: cms.RevocationInfoChoice
            if crl1.native["other_rev_info_format"] != "ocsp_response":
                logger.debug("bad ocsp data")
                return False, None
            elif crl1.native["other_rev_info"]["response_status"] != "successful":
                logger.debug(f"ocsp response status failure: {crl1.native['other_rev_info']['response_status']}")
                return False, None
            elif crl1.native["other_rev_info"]["response_bytes"]["response_type"] == "basic_ocsp_response":
                crlresp : ocsp.BasicOCSPResponse = crl1.chosen[1][1][1].parsed
                #crlresp = crl1.native["other_rev_info"]["response_bytes"]["response"]
                produced_at = crlresp["tbs_response_data"]["produced_at"].native
                cert_was_checked = False
                next_check_at = None
                for ccert in crlresp["tbs_response_data"]["responses"]:
                    if ccert["cert_id"]["serial_number"].native == cert.serial_number:
                        cert_was_checked = True
                        next_check_at = ccert["next_update"].native
                        break
                sigalgo = crlresp["signature_algorithm"]["algorithm"].native
                sig = crlresp["signature"].native
                sigok = False
                ocspcert = None
                for othercert in crlresp["certs"]:
                    for ext in othercert["tbs_certificate"]["extensions"]:
                        if ext["extn_id"].native == "extended_key_usage" and "ocsp_signing" in ext["extn_value"].native:
                            ocspcert = cx509.load_der_x509_certificate(othercert.dump())
                if ocspcert:
                    if self.validate_certificate(ocspcert, othercerts) and not sigok:
                        try:
                            public_key = ocspcert.public_key()
                            signedData = crlresp["tbs_response_data"].dump()
                            # only sha256_rsa
                            public_key.verify(
                                sig,
                                signedData,
                                padding.PKCS1v15(),
                                getattr(hashes, "SHA256")(),
                            )
                            sigok = True
                        except:
                            logger.debug(f"ocsp signing certificate is invalid")
                            pass
                if sigok and cert_was_checked:
                    return True, (produced_at, next_check_at)
                logger.debug(f"ocsp cannot be verified")
            else:
                logger.debug(f"ocsp unknown response type: {crl1.native['other_rev_info']['response_bytes']['response_type']}")
        return False, None

    def verify_tsp_data(self, signed_data, tspdata, othercerts):
        if tspdata['encap_content_info']['content_type'].native == 'tst_info':
            (_, _, tcrldata, tcert, tothercerts, _, tsignatureok) = self.decompose_signed_data(b'', tspdata)
            if not tsignatureok:
                return False, None

            def getleaf(xcert):
                leaf_certificate = None
                for cert in tothercerts:
                    if xcert.issuer == cert.subject:
                        return cert

            def seekForRoot(xcert, certs):
                xcertsigner = xcert.extensions.get_extension_for_class(
                    cx509.AuthorityKeyIdentifier
                ).value.key_identifier
                for cert in certs:
                    certsubject = cert.extensions.get_extension_for_class(
                        cx509.SubjectKeyIdentifier
                    ).value.key_identifier
                    if xcertsigner == certsubject:
                        return True
                for cert in self.certs:
                    certsubject = cert.extensions.get_extension_for_class(
                        cx509.SubjectKeyIdentifier
                    ).value.key_identifier
                    if xcertsigner == certsubject:
                        return True
                return False

            # tcert must have timeStamping EKU
            eku_extension = tcert.extensions.get_extension_for_class(
                cx509.ExtendedKeyUsage
            )
            if not eku_extension.critical:
                logger.debug("The EKU extension is not critical.")
                return False, None

            if cx509.ExtendedKeyUsageOID.TIME_STAMPING not in eku_extension.value:
                logger.debug("The EKU extension does not have KeyPurposeID id-kp-timeStamping.")
                return False, None

            leaf_certificate = getleaf(tcert)
            if leaf_certificate:
                bc_extension = leaf_certificate.extensions.get_extension_for_class(
                    cx509.BasicConstraints
                )
                eku_extension = leaf_certificate.extensions.get_extension_for_class(
                    cx509.ExtendedKeyUsage
                )
                if not bc_extension.critical or not bc_extension.value.ca:
                    logger.debug("Leaf certificate is not CA.")
                    return False, None
                #elif not eku_extension.critical:
                #    # leaf cert must have timeStamping EKU
                #    msg = "The EKU extension is not critical."
                #    raise VerificationError(msg)
                elif cx509.ExtendedKeyUsageOID.TIME_STAMPING not in eku_extension.value:
                    logger.debug("The EKU extension does not have KeyPurposeID id-kp-timeStamping.")
                    return False, None

                while True:
                    cert = getleaf(leaf_certificate)
                    if cert:
                        leaf_certificate = cert
                    else:
                        break

                # leaf_certificate must be signed by known CA
                # ok = self.verifier.verify(leaf_certificate, othercerts)
                ok = seekForRoot(leaf_certificate, othercerts)
            else:
                # tcert may not have subjectAltName
                #ok = self.verifier.verify(tcert, othercerts)
                ok = seekForRoot(tcert, othercerts)

            if not ok:
                logger.debug("No leaf certificate found in the chain.")
                return False, None

            tst = tspdata['encap_content_info']['content'].parsed
            signature_bytes = signed_data['signer_infos'][0]['signature'].native
            md = hashlib.sha256(signature_bytes).digest()
            if md == tst['message_imprint']['hashed_message'].native:
                return True, tst['gen_time'].native

        return False, None


def verify(pdfdata:bytes, certs:list[cx509.Certificate]=None) -> list[tuple[bool, bool, bool]]   :
    """
    Verify PDF signature.
    :param pdfdata: PDF document as bytes.
    :param certs: List of additional certificates used to verify signature (system independent).
    :return: List of tuples containing three boolean values for each signature:
        (hashok, signatureok, certok):
        hashok: bool - True if the hash matches.
        signatureok: bool - True if the signature is valid.
        certok: bool - True if the certificate used for signing is trusted and valid."""
    results = []
    n = pdfdata.find(b"/ByteRange")
    while n != -1:
        start = pdfdata.find(b"[", n)
        stop = pdfdata.find(b"]", start)
        assert n != -1 and start != -1 and stop != -1
        br = [int(i, 10) for i in pdfdata[start + 1 : stop].split()]
        assert pdfdata[br[1]] == 60 and pdfdata[br[2]-1] == 62
        contents = pdfdata[br[0] + br[1] + 1 : br[2] - 1]
        bcontents = bytes.fromhex(contents.decode("utf8"))
        data1 = pdfdata[br[0] : br[0] + br[1]]
        data2 = pdfdata[br[2] : br[2] + br[3]]
        signedData = data1 + data2

        result = verifier.verify(bcontents, signedData, certs)
        results.append(result)
        n = pdfdata.find(b"/ByteRange", br[2] + br[3])
    return results



# *-* coding: utf-8 *-*
from email import message_from_string

from cryptography import x509

from endesive import verifier


def verify(data:bytes, certs:list[x509.Certificate]=None) -> tuple[bool, bool, bool]:
    """
    Verifiy S/MIME signed email.

    :param data: Email data as bytes.
    :param certs: List of additional certificates used to verify signature (system independent).
    :return:
        hashok, signatureok, certok

        hashok: bool
            True if the hash matches.
        signatureok: bool
            True if the signature is valid.
        certok: bool
            True if the certificate used for signing is trusted and valid.
    """
    msg = message_from_string(data)
    sig = None
    plain = None
    for part in msg.walk():
        ct = part.get_content_type()
        # multipart/* are just containers
        if ct.split('/')[0] == 'multipart':
            continue
        if ct == 'application/x-pkcs7-signature':
            sig = part.get_payload(decode=True)
        elif ct == 'application/pkcs7-signature':
            sig = part.get_payload(decode=True)
        elif ct == 'text/plain':
            plain = part.get_payload(decode=False)
    if sig is None:
        raise ValueError('not signed email')

    plain = plain.encode('utf-8')
    plain = plain.replace(b'\n', b'\r\n')

    return verifier.verify(sig, plain, certs)



# *-* coding: utf-8 *-*
from cryptography import x509
from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes

from endesive import signer


def sign(datau:bytes, key: PrivateKeyTypes, cert: x509.Certificate, certs: list[x509.Certificate], hashalgo='sha1', attrs=True, pss=False) -> bytes:
    """
    Sign data with private key without any encapsulation.

    Parameters:
        datau: Data to sign (bytes).
        key: Private key to sign with (PrivateKeyTypes).
        cert: Certificate to sign with (x509.Certificate).
        certs: List of additional certificates (list of x509.Certificate).
        hashalgo: Hash algorithm to use (str, default 'sha1').
        attrs: Whether to include attributes (bool, default True).
        pss: Whether to use PSS padding (bool, default False).

    Returns:
        Signed data as bytes.
    """
    return signer.sign(datau, key, cert, certs, hashalgo, attrs, pss=pss)



# *-* coding: utf-8 *-*
import base64

from cryptography import x509
from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes

from endesive import signer


class SignedData(object):

    def email(self, hashalgo, datau, datas, prefix):
        s = b'''\
MIME-Version: 1.0
Content-Type: multipart/signed; protocol="application/%spkcs7-signature"; micalg="%s"; boundary="----46F1AAD10BE922477643C0A33C40D389"

This is an S/MIME signed message

------46F1AAD10BE922477643C0A33C40D389
%s
------46F1AAD10BE922477643C0A33C40D389
Content-Type: application/%spkcs7-signature; name="smime.p7s"
Content-Transfer-Encoding: base64
Content-Disposition: attachment; filename="smime.p7s"

%s
------46F1AAD10BE922477643C0A33C40D389--

''' % (prefix, hashalgo, datau, prefix, datas)
        return s

    def build(self, datau, key, cert, othercerts, hashalgo, attrs, pss=False):
        datau = datau.replace(b'\n', b'\r\n')
        datas = signer.sign(datau, key, cert, othercerts, hashalgo, attrs, pss=pss)
        datas = base64.encodebytes(datas)
        if hashalgo == 'sha1':
            hashalgo = b'sha1'
        elif hashalgo == 'sha256':
            hashalgo = b'sha-256'
        elif hashalgo == 'sha512':
            hashalgo = b'sha-512'
        prefix = [b'x-', b''][pss]
        data = self.email(hashalgo, datau, datas, prefix)
        return data


def sign(datau:bytes, key: PrivateKeyTypes, cert: x509.Certificate, certs: list[x509.Certificate], hashalgo='sha1', attrs=True, pss=False)->bytes:
    """
    Sign data with private key and encapsulate the result (data and signature) as S/MIME message.

    :param datau: Data to sign (bytes).
    :param key: Private key to sign with (PrivateKeyTypes).
    :param cert: Certificate to sign with (x509.Certificate).
    :param certs: List of additional certificates (list of x509.Certificate).
    :param hashalgo: Hash algorithm to use (str, default 'sha1').
    :param attrs: Whether to include attributes (bool, default True).
    :param pss: Whether to use PSS padding (bool, default False).
    :return: Signed data as bytes.
    """
    cls = SignedData()
    return cls.build(datau, key, cert, certs, hashalgo, attrs, pss)



# *-* coding: utf-8 *-*
from cryptography import x509

from endesive import verifier

def verify(datas:bytes, datau:bytes, certs:list[x509.Certificate]=None) -> tuple[bool, bool, bool]:
    """
    Verifies signed bytes.

    Parameters:
        datas: Signed data as bytes.
        datau: Original data as bytes.
        certs: List of additional certificates used to verify signature (system independent).

    Returns:
        hashok, signatureok, certok
        
        hashok : bool
            True if the hash matches.
        signatureok : bool
            True if the signature is valid.
        certok : bool
            True if the certificate used for signing is trusted and valid.
    """
    return verifier.verify(datas, datau, certs)



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
import datetime
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography import x509

from endesive.pdf import cms


def main():
    date = datetime.datetime.now()
    date = date.strftime("D:%Y%m%d%H%M%S+00'00'")
    dct = {
        "aligned": 16384,
        "sigflags": 3,
        "sigflagsft": 132,
        "sigpage": 0,
        # "sigbutton": True,
        # "sigfield": "Signature1",
        # "auto_sigfield": True,
        # "sigandcertify": True,
        # "signaturebox": (470, 840, 570, 640),
        "signature": "Dokument podpisany cyfrowo ąćęłńóśżź",
        # "signature_img": "signature_test.png",
        "contact": "contact:mak@trisoft.com.pl",
        "location": "Szczecin",
        "signingdate": date,
        "reason": "Dokument podpisany cyfrowo aą cć eę lł nń oó sś zż zź",
        "password": "1234",
        'tsa_url': 'http://timestamp.digicert.com',
        "ltv": True,
    }
    with open("ca/demo2_user1.p12", "rb") as fp:
        p12 = pkcs12.load_key_and_certificates(
            fp.read(), b"1234", backends.default_backend()
        )
    fname = "pdf.pdf"
    if len(sys.argv) > 2:
        fname = sys.argv[2]
    datau = open(fname, "rb").read()

    print(f"Signing certificate subject: {p12[1].subject.rfc4514_string()}")
    print(f"Signing certificate issuer: {p12[1].issuer.rfc4514_string()}")
    print(f"Additional certificates in chain: {len(p12[2]) if p12[2] else 0}")

    issuer_cert = None
    if p12[2]:
        issuer_cert = p12[2][0]
        print(
            f"Using issuer certificate: {issuer_cert.subject.rfc4514_string()}")
    else:
        print("No additional certificates found in P12 file")

    # ocsp_url = "http://ca.trisoft.com.pl/ocsp"

    datas = cms.sign(datau, dct, p12[0], p12[1], p12[2], "sha256",
                    #  ocspurl=ocsp_url,
                     ocspissuer=issuer_cert,
                     timestampurl=dct['tsa_url'])
    fname = fname.replace(".pdf", "-signed-cms-ltv.pdf")
    with open(fname, "wb") as fp:
        fp.write(datau)
        fp.write(datas)


main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
import datetime
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12

from endesive.pdf import cms

# from endesive.pdf import cmsn as cms

# import logging
# logging.basicConfig(level=logging.DEBUG)


def main():
    date = datetime.datetime.utcnow()
    date = date.strftime("D:%Y%m%d%H%M%S+00'00'")
    dct = {
        "aligned": 8192,
        "sigflags": 1,
        "sigflagsft": 132,
        "sigpage": 0,
        # "sigbutton": True,
        "sigfield": "Signature1",
        "auto_sigfield": True,
        "sigandcertify": True,
        "signaturebox": (470, 440, 570, 240),
        "signature": "Podpis 3",
        # "signature_img": "signature_test.png",
        "contact": "contact:demo3@demot.com.pl",
        "location": "Szczecin",
        "signingdate": date,
        "reason": "Dokument podpisany cyfrowo aą cć eę lł nń oó sś zż zź",
        "password": "1234",
    }
    with open("ca/demo2_user3.p12", "rb") as fp:
        p12 = pkcs12.load_key_and_certificates(
            fp.read(), b"1234", backends.default_backend()
        )
    fname = "pdf-signed-cms-twice-2.pdf"
    if len(sys.argv) > 1:
        fname = sys.argv[1]
    datau = open(fname, "rb").read()
    datas = cms.sign(datau, dct, p12[0], p12[1], p12[2], "sha256")
    fname = fname.replace("2", "end")
    with open(fname, "wb") as fp:
        fp.write(datau)
        fp.write(datas)


main()



#!/usr/bin/env vpython3
# coding: utf-8

import os
import sys
import binascii
import datetime
import PyKCS11
import base64
import hashlib

from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization

from asn1crypto import x509 as asn1x509
from asn1crypto import keys as asn1keys
from asn1crypto import pem as asn1pem
from asn1crypto import util as asn1util

import paramiko.agent

import cryptography


class BaseHSM:
    def certificate(self):
        """
        callback for HSM
        used to identfy the ssh agents key exports via fingerprint

        :return: public-key-fingerprint, certificate-in-pem
        """
        raise NotImplementedError()

    def sign(self, keyid, data, mech):
        """
        sign

        :param keyid: the keyid as returned by certificate()
        :param data:
        :param mech: hash algo
        :return: PKCS7 signature blob
        """
        raise NotImplementedError()


class HSM(BaseHSM):
    def __init__(self, dllpath):
        """
        Initialize HSM

        :param dllpath: dynamic library used to communicate with HSM
        """
        self.pkcs11 = PyKCS11.PyKCS11Lib()
        self.pkcs11.load(dllpath)
        self.session = None

    def getSlot(self, label):
        """
        Find the slot

        :param label: searched slot name
        :return: slot
        """
        slots = self.pkcs11.getSlotList(tokenPresent=True)
        for slot in slots:
            info = self.pkcs11.getTokenInfo(slot)
            try:
                if info.label.split("\0")[0].strip() == label:
                    return slot
            except AttributeError:
                continue
        return None

    def create(self, label, pin, sopin):
        """
        Initialize a slot

        :param label: searched slot name
        :param pin: pin for slot
        :param sopin: administrative pin
        """
        slot = self.getSlot(label)
        if slot is not None:
            return
        slot = self.pkcs11.getSlotList(tokenPresent=True)[-1]
        self.pkcs11.initToken(slot, sopin, label)
        session = self.pkcs11.openSession(
            slot, PyKCS11.CKF_SERIAL_SESSION | PyKCS11.CKF_RW_SESSION
        )
        session.login(sopin, user_type=PyKCS11.CKU_SO)
        session.initPin(pin)
        session.logout()
        session.closeSession()

    def login(self, label, pin):
        """
        Start session

        :param label: slot name
        :param pin: pin for slot
        """
        slot = self.getSlot(label)
        if slot is None:
            return
        self.session = self.pkcs11.openSession(
            slot, PyKCS11.CKF_SERIAL_SESSION | PyKCS11.CKF_RW_SESSION
        )
        self.session.login(pin)

    def logout(self):
        """
        End session
        """
        if self.session is not None:
            self.session.logout()
            self.session.closeSession()
            self.session = None

    def gen_privkey(self, label, key_id, key_length=2048):
        """
        Create private key

        :param label: key label
        :param key_id: key ID
        :param key_length: key length in bits
        """
        # label - just a label for identifying objects
        # key_id has to be the same for both objects, it will also be necessary
        #     when importing the certificate, to ensure it is linked with these keys.
        # key_length - key-length in bits

        public_template = [
            (PyKCS11.CKA_CLASS, PyKCS11.CKO_PUBLIC_KEY),
            (PyKCS11.CKA_TOKEN, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_PRIVATE, PyKCS11.CK_FALSE),
            (PyKCS11.CKA_MODULUS_BITS, key_length),
            #            (PyKCS11.CKA_PUBLIC_EXPONENT, (0x01, 0x00, 0x01)),
            (PyKCS11.CKA_ENCRYPT, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_VERIFY, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_VERIFY_RECOVER, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_WRAP, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_LABEL, label),
            (PyKCS11.CKA_ID, key_id)
            #            (PyKCS11.CKA_KEY_TYPE, PyKCS11.CKK_RSA),
            #            (PyKCS11.CKA_SENSITIVE, PyKCS11.CK_FALSE),
        ]

        private_template = [
            (PyKCS11.CKA_CLASS, PyKCS11.CKO_PRIVATE_KEY),
            (PyKCS11.CKA_TOKEN, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_PRIVATE, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_DECRYPT, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_SIGN, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_SIGN_RECOVER, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_UNWRAP, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_LABEL, label),
            (PyKCS11.CKA_ID, key_id)
            #            (PyKCS11.CKA_SENSITIVE, PyKCS11.CK_TRUE),
        ]

        self.session.generateKeyPair(public_template, private_template)

    def cert_save(self, cert, label, subject, key_id):
        """
        Save certificate

        :param cert: certificate
        :param label: certificate label
        :param subject: certificate subject
        :param key_id: key ID
        """
        cert_template = [
            (PyKCS11.CKA_CLASS, PyKCS11.CKO_CERTIFICATE),
            (PyKCS11.CKA_CERTIFICATE_TYPE, PyKCS11.CKC_X_509),
            (PyKCS11.CKA_TOKEN, PyKCS11.CK_TRUE),
            (PyKCS11.CKA_LABEL, label.encode("utf-8")),
            (
                PyKCS11.CKA_ID,
                key_id,
            ),  # must be set, and DER see Table 24, X.509 Certificate Object Attributes
            (
                PyKCS11.CKA_SUBJECT,
                subject.encode("utf-8"),
            ),  # must be set and DER, see Table 24, X.509 Certificate Object Attributes
            # (PyKCS11.CKA_PRIVATE, PyKCS11.CK_FALSE),
            # (PyKCS11.CKA_TRUSTED, PyKCS11.CK_TRUE),
            # (PyKCS11.CKA_SENSITIVE, PyKCS11.CK_FALSE),
            # (PyKCS11.CKA_ENCRYPT, PyKCS11.CK_TRUE),
            # (PyKCS11.CKA_VERIFY, PyKCS11.CK_TRUE),
            # (PyKCS11.CKA_MODIFIABLE, PyKCS11.CK_TRUE),
            #            (PyKCS11.CKA_ISSUER, cert.Issuer);
            #            (PyKCS11.CKA_SERIAL_NUMBER,cert.SerialNumber)
            (PyKCS11.CKA_VALUE, cert),  # must be BER-encoded
        ]

        self.session.createObject(cert_template)

    def cert_load(self, keyID):
        """
        Load certificate

        :param keyID: key ID
        """
        rec = self.session.findObjects(
            [(PyKCS11.CKA_CLASS, PyKCS11.CKO_CERTIFICATE), (PyKCS11.CKA_ID, keyID)]
        )
        if len(rec) == 0:
            return None
        value = bytes(rec[0].to_dict()["CKA_VALUE"])
        return value

    def certsign(self, sn, pubKey, subject, until, caprivKey, ca):
        """
        Sign certificate

        :param sn: serial number
        :param pubKey: public key
        :param subject: common name for certificate subject
        :param until: until when is the certificate valid
        :param caprivKey: signing key
        :param ca:

            None: create self signed root CA certificate

            True: create indirect CA certificate

            False: create user certificate
        """
        args = {
            "version": "v3",
            "serial_number": sn,
            "issuer": asn1x509.Name.build({
                "common_name": "hsm Root CA",
            }),
            "subject": asn1x509.Name.build({
                "common_name": subject,
            }),
            "signature": {
                "algorithm": "sha256_rsa",
                "parameters": None,
            },
            "validity": {
                "not_before": asn1x509.Time({
                    "utc_time": asn1x509.UTCTime((datetime.datetime.utcnow() - datetime.timedelta(days=1)).strftime('%y%m%d%H%M%SZ')),
                }),
                "not_after": asn1x509.Time({
                    "utc_time": asn1x509.UTCTime(until.strftime('%y%m%d%H%M%SZ')),
                }),
            },
            "subject_public_key_info": {
                "algorithm": {
                    "algorithm": "rsa",
                    "parameters": None,
                },
                "public_key": pubKey,
            },
        }
        if ca is None:
            args["issuer"] = asn1x509.Name.build({
                "common_name": subject,
            })
            args.update({
                "extensions": [
                    {
                        "extn_id": "basic_constraints",
                        "critical": True,
                        "extn_value": {"ca": True, "path_len_constraint": None},
                    },
                    {
                        "extn_id": "key_usage",
                        "critical": True,
                        "extn_value": set(
                            [
                                "crl_sign",
                                "digital_signature",
                                "key_cert_sign",
                            ]
                        ),
                    },
                ]
            })
        elif ca:
            extensions = [
                {
                #    'extn_id': 'crl_distribution_points',
                #    'critical': False,
                #    'extn_value': [
                #        asn1util.OrderedDict([
                #            ('distribution_point', ['http://ca.trisoft.com.pl/crl']),
                #            ('reasons', None),
                #            ('crl_issuer', None),
                #        ]),
                #    ]
                #}, {
                #    'extn_id':'authority_information_access',
                #    'critical': False,
                #    'extn_value': [
                #        {
                #            'access_method': 'ca_issuers',
                #            'access_location': asn1x509.URI('http://ca.trisoft.com.pl/cacert'),
                #        }, {
                #            'access_method': 'ocsp',
                #            'access_location': asn1x509.URI('http://ca.trisoft.com.pl/ocsp'),
                #        }
                #    ]
                #}, {
                    'extn_id': 'authority_key_identifier',
                    'critical': False,
                    'extn_value': {
                        'key_identifier': None,
                        'authority_cert_issuer': None,
                        'authority_cert_serial_number': 1,
                    }
                #}, {
                #    'extn_id': 'key_identifier',
                #    'critical': False,
                #    'extn_value': b'',
                }, {
                    "extn_id": "basic_constraints",
                    "critical": True,
                    "extn_value": {"ca": True, "path_len_constraint": 0},
                }, {
                    "extn_id": "key_usage",
                    "critical": True,
                    "extn_value": set([
                        "crl_sign",
                        "digital_signature",
                        "key_cert_sign",
                    ]),
                },
            ]
            args.update({
                "extensions": extensions
            })
        else:
            extensions = [
                #asn1util.OrderedDict([
                #    ('extn_id', 'subject_alt_name'),
                #    ('critical', False),
                #    ('extn_value', ['demo1@trisoft.com.pl'])
                #]),
                {
                #    'extn_id': 'subject_alt_name',
                #    'critical': False,
                #    'extn_value': {'rfc822_name' : 'demo1@trisoft.com.pl'},
                #}, {
                    'extn_id': 'authority_key_identifier',
                    'critical': False,
                    'extn_value': {
                        'key_identifier': None,
                        'authority_cert_issuer': None,
                        'authority_cert_serial_number': 2,
                    }
                #}, {
                #    'extn_id': 'key_identifier',
                #    'critical': False,
                #    'extn_value': b'',
                }, {
                    "extn_id": "basic_constraints",
                    "critical": True,
                    "extn_value": {"ca": False},
                }, {
                    "extn_id": "key_usage",
                    "critical": True,
                    "extn_value": set([
                            "digital_signature",
                            "key_agreement",
                            "key_encipherment",
                            "non_repudiation",
                    ]),
                },
            ]
            args.update({
                "issuer": asn1x509.Name.build({
                    "common_name": "hsm Indirect CA",
                }),
                "extensions": extensions,
            })

        tbs = asn1x509.TbsCertificate(args)

        # Sign the TBS Certificate
        data = tbs.dump()
        value = self.session.sign(
            caprivKey, data, PyKCS11.Mechanism(PyKCS11.CKM_SHA256_RSA_PKCS, None)
        )
        value = bytes(bytearray(value))

        cert = asn1x509.Certificate(
            {
                "tbs_certificate": tbs,
                "signature_algorithm": {
                    "algorithm": "sha256_rsa",
                    "parameters": None,
                },
                "signature_value": value,
            }
        )
        return cert.dump()

    def ca_gen(self, label, keyID, subject):
        """
        Initiate root CA certificate

        :param label: HSM key label
        :param keyID: HSM key ID
        :param subject: common name in certificate subject
        """
        privKey = self.session.findObjects(
            [(PyKCS11.CKA_CLASS, PyKCS11.CKO_PRIVATE_KEY), (PyKCS11.CKA_ID, keyID)]
        )[0]
        pubKey = self.session.findObjects(
            [(PyKCS11.CKA_CLASS, PyKCS11.CKO_PUBLIC_KEY), (PyKCS11.CKA_ID, keyID)]
        )[0]

        modulus = self.session.getAttributeValue(pubKey, [PyKCS11.CKA_MODULUS])[0]
        modulus = binascii.hexlify(bytearray(modulus)).decode("utf-8")
        exponent = self.session.getAttributeValue(
            pubKey, [PyKCS11.CKA_PUBLIC_EXPONENT]
        )[0]
        exponent = binascii.hexlify(bytearray(exponent)).decode("utf-8")
        pubKey = asn1keys.RSAPublicKey(
            {
                "modulus": int("0x" + modulus, 16),
                "public_exponent": int("0x" + exponent, 16),
            }
        )
        # pubKey = asn1keys.RSAPublicKey.load(pubKey.dump())
        until = datetime.datetime.utcnow() + datetime.timedelta(
            days=365 * 40
        )
        der_bytes = self.certsign(1, pubKey, subject, until, privKey, None)
        self.cert_save(der_bytes, label, subject, keyID)

    def ca_sign(self, keyID, label, sn, subject, days, cakeyID):
        """
        Sign certificate

        :param keyID: HSM key identifier of the signed certificate
        :param label: HSM label
        :param sn: certificate serial number
        :param subject: HSM subject, common name stored in subject fiel of the signed certificate
        :param days: validity day of certificate
        :param cakeyID: HSM key identifier of the signing key
        """
        caprivKey = self.session.findObjects(
            [(PyKCS11.CKA_CLASS, PyKCS11.CKO_PRIVATE_KEY), (PyKCS11.CKA_ID, cakeyID)]
        )[0]

        pubKey = self.session.findObjects(
            [(PyKCS11.CKA_CLASS, PyKCS11.CKO_PUBLIC_KEY), (PyKCS11.CKA_ID, keyID)]
        )[0]
        modulus = self.session.getAttributeValue(pubKey, [PyKCS11.CKA_MODULUS])[0]
        modulus = binascii.hexlify(bytearray(modulus)).decode("utf-8")
        exponent = self.session.getAttributeValue(
            pubKey, [PyKCS11.CKA_PUBLIC_EXPONENT]
        )[0]
        exponent = binascii.hexlify(bytearray(exponent)).decode("utf-8")
        pubKey = asn1keys.RSAPublicKey(
            {
                "modulus": int("0x" + modulus, 16),
                "public_exponent": int("0x" + exponent, 16),
            }
        )
        # pubKey = asn1keys.RSAPublicKey.load(pubKey.dump())
        until = datetime.datetime.utcnow() + datetime.timedelta(
            days=days
        )
        der_bytes = self.certsign(sn, pubKey, subject, until, caprivKey, keyID == b'\x02')
        self.cert_save(der_bytes, label, subject, keyID)

    def cert_export(self, fname, keyID):
        der_bytes = self.cert_load(keyID)
        pem_bytes = asn1pem.armor("CERTIFICATE", der_bytes)
        with open(fname + ".der", "wb") as fp:
            fp.write(der_bytes)
        with open(fname + ".pem", "wb") as fp:
            fp.write(pem_bytes)


class SSHAgentHSM(BaseHSM):
    def __init__(self, cert):
        assert isinstance(cert, cryptography.x509.Certificate)
        self._a = paramiko.agent.Agent()
        self._cert = cert

    def close(self):
        self._a.close()

    def certificate(self):
        """
        callback for HSM
        used to identfy the ssh agents key exports via fingerprint

        :return: public-key-fingerprint, certificate-in-pem
        """

        # https://superuser.com/questions/421997/what-is-a-ssh-key-fingerprint-and-how-is-it-generated
        # convert RSA Key to SSH Fingerprint
        alg, key = (
            self._cert.public_key()
            .public_bytes(
                encoding=cryptography.hazmat.primitives.serialization.Encoding.OpenSSH,
                format=cryptography.hazmat.primitives.serialization.PublicFormat.OpenSSH,
            )
            .split(b" ")
        )

        fp = b"SHA256:" + base64.b64encode(
            hashlib.sha256(base64.b64decode(key)).digest()
        )
        cert = self._cert.public_bytes(
            cryptography.hazmat.primitives.serialization.Encoding.PEM
        )

        return fp, cert

    @staticmethod
    def _decode_fp(keyfp):
        """
        decode a fingerprint

        :param keyfp: key fingerprint in OpenSSH Format
        :return: alg, fingerprint-binary
        """
        if not isinstance(keyfp, str):
            keyfp = keyfp.decode()
        alg, other = keyfp.split(":", 1)
        if alg == "SHA256":
            # pad base64 data
            data = other.encode() + b"=" * (-len(other) % 4)
            fp = base64.b64decode(data)
        elif alg == "MD5":
            data = other.replace(":", " ")
            fp = bytes.fromhex(data)
        else:
            raise ValueError(alg)
        return alg.lower(), fp

    def key(self, fp):
        """
        lookup a ssh-agent-exported key using fingerprint

        :param fp: the fingerprint
        :return: the key on success
        """

        alg, fp = self._decode_fp(fp)
        for key in self._a.get_keys():
            kfp = getattr(hashlib, alg)(key.asbytes()).digest()
            if kfp == fp:
                break
        else:
            raise ValueError("Key not found")
        return key

    def sign(self, keyid, data, hashalgo):
        """
        sign using ssh-agent sign_data
        creates RSA signature with padding=PKCS1v15 alg=SHA1

        :param keyid: the keyid as returned by certificate()
        :param data:
        :param hashalgo: has to be sha1, sha256 or sha512
        :return: PKCS7 signature blob
        """
        assert hashalgo in ("sha1", "sha256", "sha512")

        if not isinstance(data, bytes):
            data = data.encode()

        # defined in
        # SSH Agent Protocol draft-miller-ssh-agent-00 5.3.  Signature flags
        # https://tools.ietf.org/html/draft-miller-ssh-agent-00#section-5.3
        flags = {
            "sha1": 0,
            "sha256": 2,  # SSH_AGENT_RSA_SHA2_256
            "sha512": 4,  # SSH_AGENT_RSA_SHA2_512
        }[hashalgo]

        key = self.key(keyid)

        # AgentKey.sign_ssh_data is padding=PKCS1v15 alg=SHA1
        # paramiko does not expose the ssh-agent sign flags to use sha2-256/512
        # re-implement sign_ssh_agent ..
        msg = paramiko.message.Message()
        msg.add_byte(paramiko.agent.cSSH2_AGENTC_SIGN_REQUEST)
        msg.add_string(key.blob)
        msg.add_string(data)
        msg.add_int(flags)
        ptype, result = self._a._send_message(msg)
        if ptype != paramiko.agent.SSH2_AGENT_SIGN_RESPONSE:
            raise paramiko.SSHException("key cannot be used for signing")
        d = paramiko.message.Message(result.get_binary())

        # parse operation result
        alg = d.get_text()

        # interpret
        if alg in ("ssh-rsa", "rsa-sha2-256", "rsa-sha2-512"):
            sig = d.get_binary()
        else:
            raise ValueError(alg)
        return sig



import sys

if sys.platform == "win32":
    dllpath = r"c:\windows\system32\cryptoCertum3PKCS.dll"
else:
    import ctypes as ct
    if 0:
        openssl_1_1 = [
            ct.CDLL('/usr/lib/x86_64-linux-gnu/libcrypto.so', ct.RTLD_GLOBAL),
            ct.CDLL('/usr/lib/x86_64-linux-gnu/libssl.so', ct.RTLD_GLOBAL),
        ]
    dllpath = '/devel/lib/pkcs11libs/libCCGraphiteP11.2.0.5.6.so'
